BorderedMorph subclass: #ElementsScrollFrameMorph
	instanceVariableNames: 'contents hScrollbar vScrollbar cornerMorph hbarInset vbarInset contentsChanged growthFraction '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Elements-UI'!
!ElementsScrollFrameMorph commentStamp: '<historical>' prior: 0!
I have largely been copied from the Scratch-UI which was originally written by John Maloney and Evelyn Eastmond!


!ElementsScrollFrameMorph methodsFor: 'initialization' stamp: 'jens 2/4/2009 20:53'!
initialize

	super initialize.
	self color: Color red.
	hbarInset _ 12.
	vbarInset _ 12.
	contentsChanged _ false.
	growthFraction _ 0.6.  "amount to grow when contents do not fit as a fraction of the current size"

	contents _ PasteUpMorph new
		color: Color red;
		borderWidth: 0;
		enableDragNDrop: true.

	hScrollbar _ ScrollBar new
		model: self;
		setValueSelector: #hScrollRelative:;
		hasMenuButton: false.
	vScrollbar _ ScrollBar new
		model: self;
		setValueSelector: #vScrollRelative:;
		hasMenuButton: false.

	cornerMorph _ BorderedMorph new
		borderRaised; borderWidth: 1; color: hScrollbar color.

	contents position: self position + borderWidth.
	self addMorph: contents.
	self addMorph: hScrollbar.
	self addMorph: vScrollbar.
	self addMorph: cornerMorph.
	self extent: 160@120.
! !


!ElementsScrollFrameMorph methodsFor: 'accessing' stamp: 'jens 2/4/2009 20:28'!
contents
	"Answer my contents morph."

	^ contents
! !

!ElementsScrollFrameMorph methodsFor: 'accessing' stamp: 'jens 2/4/2009 20:28'!
contents: aMorph
	"Replace my contents morph. The argument can be any morph. Typically it is a BorderedMorph or a PasteUpMorph."

	contents ifNotNil: [contents delete].
	contents _ aMorph.
	contents position: self topLeft + borderWidth.
	self addMorphBack: contents.
	self extent: self extent.
! !

!ElementsScrollFrameMorph methodsFor: 'accessing' stamp: 'jens 2/4/2009 20:28'!
growthFraction: aNumber
	"Set the amount to grow when contents do not fit, expressed as a fraction of the current size. For example, 0.5 makes the contains pane 50% larger if any submorph extends over the edge."

	growthFraction _ aNumber max: 0.0.
! !

!ElementsScrollFrameMorph methodsFor: 'accessing' stamp: 'jens 2/4/2009 20:28'!
hBarInset: anInteger

	hbarInset _ anInteger.
! !

!ElementsScrollFrameMorph methodsFor: 'accessing' stamp: 'jens 2/4/2009 20:28'!
vBarInset: anInteger

	vbarInset _ anInteger.
! !


!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
hScrollPixels
	"Answer the current horizontal scroll offset in pixels."

	^ (self left + borderWidth) - contents left

! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
hScrollPixels: hOffset
	"Scroll to the given horizontal offset in pixels. Zero is scrolled to the left and increasing values scroll right."

	| delta maxOffset |
	delta _ (hOffset asInteger min: self maxScroll x) max: 0.
	contents left: ((self left + borderWidth) - delta) truncated.

	maxOffset _ self maxScroll x.
	maxOffset > 0 ifTrue: [hScrollbar value: self hScrollPixels / maxOffset].
! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
hScrollRelative: aFloat
	"Sent by the horizontal scrollbar. Scroll to the given relative postion between 0.0 and 1.0."

	self hScrollPixels: aFloat * self maxScroll x.
! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
maxScroll
	"Answer a point representing the maximum horizontal and vertical scroll offsets in pixels."

	^ contents extent - self visibleExtent
! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
scrollSubmorphIntoView: aMorph

	| offset |
	(contents submorphs includes: aMorph) ifFalse: [^ self].
	(self bounds containsRect: aMorph fullBounds) ifTrue: [^ self].

	(aMorph fullBounds bottom > self bottom) ifTrue: [
		offset _ aMorph fullBounds bottom - contents top.
		contents top: (self bottom - 3) - offset].

	(aMorph fullBounds top < self top) ifTrue: [
		offset _ aMorph fullBounds top - contents top.
		contents top: (self top + 3) - offset].

	self updateScrollbars.
! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
vScrollPixels
	"Answer the current vertical scroll offset in pixels."

	^ (self top + borderWidth) - contents top

! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
vScrollPixels: vOffset
	"Scroll to the given vertical offset in pixels. Zero is scrolled to the top and increasing values scroll down."

	| delta maxOffset |
	delta _ (vOffset asInteger min: self maxScroll y) max: 0.
	contents top: ((self top + borderWidth) - delta) truncated.

	maxOffset _ self maxScroll y.
	maxOffset > 0 ifTrue: [vScrollbar value: self vScrollPixels / maxOffset].
! !

!ElementsScrollFrameMorph methodsFor: 'scrolling' stamp: 'jens 2/4/2009 20:28'!
vScrollRelative: aFloat
	"Sent by the vertical scrollbar. Scroll to the given relative postion between 0.0 and 1.0."

	self vScrollPixels: aFloat * self maxScroll y.
! !


!ElementsScrollFrameMorph methodsFor: 'scrollbar visibility' stamp: 'jens 2/4/2009 20:28'!
showHorizontalScrollbar: aBoolean
	"Show or hide my horizontal scrollbar."

	aBoolean
		ifTrue: [
			self addMorph: hScrollbar.
			vScrollbar owner = self ifTrue: [self addMorph: cornerMorph]]
		ifFalse: [
			hScrollbar delete.
			cornerMorph delete].

	super extent: self extent.
! !

!ElementsScrollFrameMorph methodsFor: 'scrollbar visibility' stamp: 'jens 2/4/2009 20:28'!
showVerticalScrollbar: aBoolean
	"Show or hide my vertical scrollbar."

	aBoolean
		ifTrue: [
			self addMorph: vScrollbar.
			hScrollbar owner = self ifTrue: [self addMorph: cornerMorph]]
		ifFalse: [
			vScrollbar delete.
			cornerMorph delete].

	super extent: self extent.
! !


!ElementsScrollFrameMorph methodsFor: 'drawing' stamp: 'jens 2/4/2009 20:28'!
drawSubmorphsOn: aCanvas
	"If my contents has changed, fix it's extent and update my scrollbar ranges. Clip submorph drawing to my bounds."

	contentsChanged ifTrue: [
		self updateContentsExtent.
		self updateScrollbars.
		contentsChanged _ false].

	super drawSubmorphsOn: (aCanvas copyClipRect: self innerBounds).
! !


!ElementsScrollFrameMorph methodsFor: 'layout' stamp: 'jens 2/4/2009 20:28'!
hResizing

	^ #spaceFill
! !

!ElementsScrollFrameMorph methodsFor: 'layout' stamp: 'jens 2/4/2009 20:28'!
isAlignmentMorph
	"Answer true, since I can be laid out as if I were an AlignmentMorph. I pretend to be an AlignmentMorph so that I can be resized when I'm inside another AlignmentMorph."

	^ true
! !

!ElementsScrollFrameMorph methodsFor: 'layout' stamp: 'jens 2/4/2009 20:28'!
layoutInWidth: w height: h
	"Resize myself to the given width and height. Called during when I am in an AlignmentMorph."

	self extent: w@h.
! !

!ElementsScrollFrameMorph methodsFor: 'layout' stamp: 'jens 2/4/2009 20:28'!
minHeight

	^ 45
! !

!ElementsScrollFrameMorph methodsFor: 'layout' stamp: 'jens 2/4/2009 20:28'!
minWidth

	^ 60
! !

!ElementsScrollFrameMorph methodsFor: 'layout' stamp: 'jens 2/4/2009 20:28'!
vResizing

	^ #spaceFill
! !


!ElementsScrollFrameMorph methodsFor: 'geometry' stamp: 'jens 2/4/2009 20:28'!
extent: aPoint
	"After setting my size, position and size my scrollbars and grow box. Also update my contents and scrollbar ranges."

	| inner w h |
	super extent: (aPoint truncated max: self minWidth@self minHeight).
	inner _ self innerBounds.

	w _ inner width.
	vScrollbar owner = self ifTrue: [w _ w - vbarInset].
	hScrollbar position: inner left@(inner bottom - hbarInset).
	hScrollbar extent: w@hbarInset.

	h _ inner height.
	hScrollbar owner = self ifTrue: [h _ h - hbarInset].
	vScrollbar position: (inner right - vbarInset)@inner top.
	vScrollbar extent: vbarInset@h.

	cornerMorph position: (inner bottomRight - (vbarInset@hbarInset)).
	cornerMorph extent: vbarInset@hbarInset.

	self updateContentsExtent.
	self updateScrollbars.
! !

!ElementsScrollFrameMorph methodsFor: 'geometry' stamp: 'jens 2/4/2009 20:28'!
fullBounds
	"Overridden to clip submorph hit detection to my bounds."

	^ bounds
! !

!ElementsScrollFrameMorph methodsFor: 'geometry' stamp: 'jens 2/4/2009 20:28'!
invalidRect: damageRect
	"Clip damage reports to my bounds, since drawing is clipped to my bounds."

	| r |
	r _ damageRect intersect: self bounds.
	(r width > 0 and: [r height > 0]) ifTrue: [super invalidRect: r].
! !

!ElementsScrollFrameMorph methodsFor: 'geometry' stamp: 'jens 2/4/2009 20:28'!
layoutChanged
	"If my contents morph's layout has changed, record that fact."

	super layoutChanged.
	contents mayNeedLayout ifTrue: [contentsChanged _ true].
! !


!ElementsScrollFrameMorph methodsFor: 'private' stamp: 'jens 2/4/2009 20:28'!
contentsBounds
	"Answer a rectangle that encloses the bounds of all my submorphs."

	| offset r |
	offset _ contents topLeft negated.
	r _ 0@0 extent: 1@1.
	contents submorphsDo: [:m |
		r _ r quickMerge: (m fullBounds translateBy: offset) truncated].
	^ r
! !

!ElementsScrollFrameMorph methodsFor: 'private' stamp: 'jens 2/5/2009 01:28'!
updateContentsExtent
	"Make sure my content morph is large enough to both enclose all it's submorphs and to fill me. Adjust the scroll offsets if my bounds have shrunk."

	| r delta scrolledOff w h visible |
	r _ self contentsBounds.
	((r left < 0) or: [r top < 0]) ifTrue: [
		"contents has morphs that stick out over its top or left edge"
		delta _ (r topLeft truncated min: (0@0)) negated.
		contents submorphsDo: [:m | m position: m position + delta]].  "move all submorphs by delta"

	visible _ self visibleExtent - (8@20).
	w _ visible x. r width > w ifTrue: [w _ (r width + (growthFraction * visible x)) rounded].
	h _ visible y. r height > h ifTrue: [h _ (r height + (growthFraction * visible y))  rounded].

	scrolledOff _ ((self topLeft + borderWidth) - contents position) max: 0@0.
	contents extent: ((w@h) max: (self visibleExtent + scrolledOff)).
	contentsChanged _ false.
! !

!ElementsScrollFrameMorph methodsFor: 'private' stamp: 'jens 2/4/2009 20:28'!
updateScrollbars
	"Update my scrollbars based on my current contents."

	| visibleExtent currentScroll maxScroll |
	contents ifNil: [^ self].  "not fully initialized yet"
	visibleExtent _ self visibleExtent.
	hScrollbar percentVisible: (visibleExtent x asFloat / contents width).
	vScrollbar percentVisible: (visibleExtent y asFloat / contents height).

	currentScroll _ (self topLeft + borderWidth) - contents position.
	contents width > visibleExtent x
		ifTrue: [
			maxScroll _ contents width - visibleExtent x.
			self showHorizontalScrollbar: true.
			hScrollbar
				scrollDelta: 4 / maxScroll asFloat 
				pageDelta: (visibleExtent x - 40) / maxScroll asFloat.
			hScrollbar value: currentScroll x / maxScroll]
		ifFalse: [
			self showHorizontalScrollbar: false.
			hScrollbar value: 0].

	contents height > visibleExtent y
		ifTrue: [
			maxScroll _ contents height - visibleExtent y.
			self showVerticalScrollbar: true.
			vScrollbar
				scrollDelta: 4.0 / maxScroll asFloat 
				pageDelta: (visibleExtent y - 40) / maxScroll asFloat.
			vScrollbar value: currentScroll y / maxScroll]
		ifFalse: [
			self showVerticalScrollbar: false.
			vScrollbar value: 0].
! !

!ElementsScrollFrameMorph methodsFor: 'private' stamp: 'jens 2/4/2009 20:28'!
visibleExtent
	"Answer the extent of my visible area. That is, the area within my borders minus the space used for scrollbars."

	| clipToScrollBarEdge visibleW visibleH |
	clipToScrollBarEdge _ false.
	clipToScrollBarEdge ifFalse: [^ self extent - (2 * borderWidth)].

	visibleW _ self width - (2 * borderWidth).
	vScrollbar owner = self ifTrue: [visibleW _ visibleW - vbarInset].
	visibleH _ self height - (2 * borderWidth).
	hScrollbar owner = self ifTrue: [visibleH _ visibleH - hbarInset].
	^ visibleW@visibleH
! !
