package enchanting;

import java.awt.Rectangle;
import java.io.*;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;
import javax.microedition.lcdui.Font;
import lejos.nxt.LCD;

public class EncSprite {
	float x, y;
	float rotationDegrees = 0.0f;
	int currentCostume = 0;		// Start invalid.  Valid numbers are from 1 to the number of costumes.
	String filename;
	int numCostumes;
	Costume costume = new Costume();
	boolean need_to_redraw;
	String message = "";
	int[] costumeOffset;
	String[] costumeName;
	boolean validData = false;
	int msgBubbleXOffset = 10;
	int msgBubbleYOffset = -10;
	boolean hidden = false;
	boolean penDown = false;
	int prevX = 0;
	int prevY = 0;
	int penColour = 0x00000000;
	//Costume[] costumes;
	//int transform = Sprite.TRANS_NONE; //Sprite.TRANS_MIRROR;
	
	static EncSprite stage = null;
	static StageOutputStream stage_out = new StageOutputStream();
	public static PrintStream out = new PrintStream(stage_out);
	
	static int X_ORIGIN = LCD.SCREEN_WIDTH / 2;
	static int Y_ORIGIN = LCD.SCREEN_HEIGHT / 2;
			
	public EncSprite(float x, float y, String filename, 
		int numCostumes, int currentCostume, float direction, boolean hidden) {
		this.x = x;
		this.y = y;
		this.filename = filename;
		if (filename.startsWith("stage")) { stage = this; }
		this.numCostumes = numCostumes;
		setRotationDegrees(direction);
		this.hidden = hidden;
		readHeaderInfoFromFile();
		wearCostume(currentCostume);
	}
	
	private void readHeaderInfoFromFile() {
		this.validData = false;

		String failPos = "A";  // debugging position

		try {
			DataInputStream in = new DataInputStream(
				new FileInputStream(new File(this.filename)));
			
			failPos = "B";
			
			// The first four bytes should be a sentinel value -- SPR0
			if (in.readInt() != 0x53505230) {
				say("Bad sentinel");
				return;
			}
			
			failPos = "C";

			// Next two bytes are the number of costumes
			if (in.readShort() != this.numCostumes) {
				say("Bad # costumes");
				return;
			}
			
			failPos = "D";
			
			// Then we have a table of offsets into the file
			// to get at the image data
			costumeOffset = new int[numCostumes];
			for (int i = 0; i < numCostumes; ++i) {
				costumeOffset[i] = in.readInt();
			}
			
			failPos = "E";

			// and a table of names of the costumes
			costumeName = new String[numCostumes];
			for (int i = 0; i < numCostumes; ++i) {
				costumeName[i] = in.readUTF();
			}
			
			failPos = "F";
			
			this.validData = true;
			
			// After the header is the graphical data, which we don't read in at this point	
					
		} catch (IOException e) {
			//say(e.getMessage());
			say("Load failed: " + failPos);
		}
	}
	
	static public Graphics getStageGraphics() {
		if (stage == null) return null;
		Image i = stage.costume.getImage();
		if (i == null) return null;
		return i.getGraphics();
	}
			
	// Set the costume that we are going to display
	public void wearCostume(int costumeNumber) {
	
		if (!this.validData) return;
	
		// In Scratch, an invalid index number has no effect on the costume
		if (costumeNumber < 1 || costumeNumber > numCostumes) return;
		
		// If we are already wearing the desired costume, do nothing
		if (costumeNumber == currentCostume) return;
		
		currentCostume = costumeNumber;
		costume.loadCostume(this.filename, costumeOffset[currentCostume - 1]);
				
		msgBubbleXOffset = costume.getBubbleXOffset();
		msgBubbleYOffset = costume.getBubbleYOffset();		
		
		if (this == stage) {
			stage_out.setImage(costume.getImage());
		}
		
		need_to_redraw = true;
	}
	
	public void wearCostume(String nameOfCostume) {
		if (!this.validData) return;
		
		// Go through our array of costume names and find this one
		int i;
		for (i = 0; i < numCostumes; ++i) {
			if (nameOfCostume.equalsIgnoreCase(costumeName[i])) {
				wearCostume(i + 1);
				break;
			}
		}
		
		if (i == numCostumes) {
			// We didn't find the name; try interpreting the string as a number
			try {
				int n = (int)(Float.parseFloat(nameOfCostume) + 0.5f);
				wearCostume(n);
        	} catch (NumberFormatException e) {
				// nothing more we can do
        	}
		}
	}
	
	public int getCurrentCostume() {
		return currentCostume;
	}
	
	public void nextCostume() {
		int next = currentCostume + 1;
		if (next > numCostumes) next = 1;
		wearCostume(next);
	}
	
	// return my bounding box (in screen space!)
	public Rectangle getBounds() {
		int spriteX = X_ORIGIN + (int)x;
		int spriteY = Y_ORIGIN - (int)y;
		return costume.getBounds(spriteX, spriteY);		
	}
	
	// draw the current sprite, and anything we are saying, on the screen
	public void drawOn(Graphics g, boolean markAsDrawn) {
	
		if (markAsDrawn) {
			need_to_redraw = false;
		}
	
		if (hidden) return;
	
		int spriteX = X_ORIGIN + (int)x;
		int spriteY = Y_ORIGIN - (int)y;
		if (this.validData) {
			costume.drawOn(g, spriteX, spriteY);
		}

				
		// Draw a line to show the direction we are facing
		/*
		float length = 15.0f;
		float radians = (float)Math.toRadians(this.rotationDegrees);
		float dX = (float)Math.cos(radians) * length;
		float dY = -(float)Math.sin(radians) * length;
		g.drawLine(spriteX, spriteY, (int)(spriteX + dX), (int)(spriteY - dY));
		*/
	}
	
	public void drawSpeechBubbleOn(Graphics g) {
		
		// Draw any text that we are showing
		if (message != null && message.length() > 0) {

			int spriteX = X_ORIGIN + (int)x;
			int spriteY = Y_ORIGIN - (int)y;

			// left, top, right, and bottom are the bounds of the text

			Font font = g.getFont();
			int width = font.stringWidth(message);
			int height = font.getHeight();

			int left = spriteX + msgBubbleXOffset;
			int top  = spriteY + msgBubbleYOffset - height;
			int right = left + width;
			int bottom = top + height;
						
			// adjust to stay on screen
			int x_adj = 0, y_adj = 0;
			if (right > LCD.SCREEN_WIDTH) {
				x_adj = LCD.SCREEN_WIDTH - right;
			}
			if (left < 0) {
				x_adj = 0 - left;
			}
			if (bottom > LCD.SCREEN_HEIGHT) {
				y_adj = LCD.SCREEN_HEIGHT - bottom;
			}
			if (top < 0) {
				y_adj = 0 - top;
			}
			left += x_adj; right += x_adj;
			top += y_adj; bottom += y_adj;
			
			// extended bounds (the area we draw in white)
			int eleft = left - 1;
			int etop = top - 1;
			int ebottom = bottom + 1;
			int eright = right + 0;
			int ewidth = eright - eleft;
			int eheight = ebottom - etop;			
			
			// Draw a white background
			g.setColor(0x00FFFFFF);
			g.fillRect(eleft, etop, ewidth, eheight);
						
			// Draw a black outline (slightly rounded)
			g.setColor(0x00000000);
			
			eright -= 1;
			ebottom -= 1;
			
			g.drawLine(eleft, etop    - 1, eright, etop    - 1);
			g.drawLine(eleft, ebottom + 1, eright, ebottom + 1);
			g.drawLine(eleft  - 1, etop, eleft  - 1, ebottom);
			g.drawLine(eright + 1, etop, eright + 1, ebottom);		
				
			g.drawString(message, left, top, Graphics.LEFT | Graphics.TOP, false);
			
			// To Do: draw an arrow sort of thing to make this a speech buble
			// or circles to make it a thought bubble.
			
			/*
			// draw a line pointing to the sprite
			g.drawLine(eleft, ebottom + 8, eleft + 8, ebottom);

			g.setColor(0x00FFFFFF);
			g.drawLine(eleft + 1, ebottom + 8, eleft + 9, ebottom);
			g.setColor(0x00000000);
			*/
		}		
	}
	
	public boolean needsRedraw() {
		return need_to_redraw;
	}
	
	// Is our sprite touching the edge (or even fully outside of bounds?)
	private boolean touchingEdge() {
		Rectangle stage_rect = new Rectangle(0, 0, LCD.SCREEN_WIDTH, LCD.SCREEN_HEIGHT);		
		return !stage_rect.contains(getBounds());
	}
	
	// Are two sprites intersecting or overlapping?
	public boolean touching(EncSprite other) {
		
		// If other is null, we are testing against the edge
		if (other == null) { return touchingEdge(); }
		
		// An sprite that isn't onscreen never collides
		if (!this.validData || !other.validData 
			|| this.hidden || other.hidden
			|| this.costume.isInvalid() || other.costume.isInvalid()) {
			return false;
		}
		
		Rectangle myRect = this.getBounds();
		Rectangle otherRect = other.getBounds();
		
		// First pass result
		if (myRect.intersects(otherRect))
		{
			// We will draw an image where the two sprites overlap
			// Just drawing the masks and or them together.
			// (The masks' image is white, and its stencil is black).
			// If there is no overlap, everything will be black.

			// Get the intersecting rectangle
			// Note I am assuming that the width and height are always positive.			
			int maxLeft   = Math.max(myRect.x, otherRect.x);
			int maxTop    = Math.max(myRect.y, otherRect.y);
			int minRight  = Math.min(myRect.x + myRect.width, otherRect.x + otherRect.width);
			int minBottom = Math.min(myRect.y + myRect.height, otherRect.y + otherRect.height);

			int width  = minRight - maxLeft;
			int height = minBottom - maxTop;

			// Uncomment the lines starting with three slashes to see a visual representation
			// of how this routine works.			 
			/// // Draw a rectangle showing the intersecting area
			/// Graphics screen = new Graphics();
			/// screen.drawRect(maxLeft, maxTop, width, height);
			 
			// Create an image large enough to hold this area
			Image i = Image.createImage(width, height);
			Graphics g = i.getGraphics();
			g.translate(-maxLeft, -maxTop);
			
			this.costume.drawMaskOn(g, X_ORIGIN + (int)this.x, Y_ORIGIN - (int)this.y, LCD.ROP_OR);
			other.costume.drawMaskOn(g, X_ORIGIN + (int)other.x, Y_ORIGIN - (int)other.y, LCD.ROP_OR);

			/// screen.drawImage(i, 0, 0, Graphics.LEFT | Graphics.TOP);

			// If there are any non-black pixels, we are intersecting
			return !imageIsColor(i, 0);
		}
		return false;
	}
	
	// Is the sprite touching a specific color?
	// Parameters: stageColorList - the color in the background we are looking for
	//             myColorList - if non-nil and valid, only check areas where our sprite is this color
	//			   all_sprites - a list of all sprites to draw (in z-order) 
	public boolean isTouchingColor(EObject stageColorList, EObject myColorList, EncSprite[] all_sprites) {
		
		// To understand how this works, it is useful to know how the images are stored.
		// The mask is black where the image is transparent and white where it is opaque.
		// The image is white where it is transparent, and black or white where it is opaque.
		
		if (!this.validData || this.hidden || this.costume.isInvalid()) return false;	// nothing to touch with
		
		int stageColorNum = colorListToNumber(stageColorList);
		if (stageColorNum == -1) return false;	// can't be touching a non-existent color
		
		int myColorNum = colorListToNumber(myColorList);
		
		int mask_rop  = LCD.ROP_NOOP;
		int image_rop = LCD.ROP_NOOP;
		int color_to_check_for = 0; 
		
		// First, do we touch the specified background color at all?			
		if (stageColorNum == 0) {
			// Show the background through transparent areas in the mask, leaving all else white.
			mask_rop = LCD.ROP_ANDINVERTED;	
			// Check to see if it is all white; if so, not touching.
			color_to_check_for = 1;		
						
			// Are we concerned with the foreground color that is touching?
			if (myColorNum == 0) {
				image_rop = LCD.ROP_AND;
			} else if (myColorNum == 1) {
				image_rop = LCD.ROP_ANDINVERTED;
			}			
		} else {
			// Show the background through transparent areas in the mask, leaving all else white.
			mask_rop = LCD.ROP_OR;
			// Check to see if all is black; if so, not touching.
			color_to_check_for = 0;
			
			// Are we concerned with the foreground color that is touching?
			if (myColorNum == 0) {
				image_rop = LCD.ROP_ORINVERTED;
			} else if (myColorNum == 1) {
				image_rop = LCD.ROP_OR;
			}
		}
		
		
		Rectangle rect = this.getBounds();
		
		// Be sure the rectangle doesn't extend beyond the edges of the screen
		int left   = Math.max(rect.x, 0);
		int top    = Math.max(rect.y, 0);
		int right  = Math.min(rect.x + rect.width,  LCD.SCREEN_WIDTH);
		int bottom = Math.min(rect.y + rect.height, LCD.SCREEN_HEIGHT);
		
		rect.setBounds(left, top, right - left, bottom - top);
		if (rect.width <= 0 || rect.height <= 0) return false;	// not touching; totally off screen

		// Create an image large enough to hold this area
		Image i = Image.createImage(rect.width, rect.height);
		Graphics g = i.getGraphics();
		g.translate(-rect.x, -rect.y);
		
		// Drawing all sprites except this one
		for (EncSprite sprite : all_sprites) {
			if (sprite != this) {
				sprite.drawOn(g, false);
			}
		}
			
		this.costume.drawMaskOn(g, X_ORIGIN + (int)this.x, Y_ORIGIN - (int)this.y, mask_rop);

		if (image_rop != LCD.ROP_NOOP) {
			this.costume.drawImageOn(g, X_ORIGIN + (int)this.x, Y_ORIGIN - (int)this.y, image_rop);
		}


		// Uncomment the lines starting with three slashes to see a visual representation
		// of how this routine works.			 
		/// Graphics stage = all_sprites[0].costume.getImage().getGraphics();
		/// stage.drawImage(i, 0, 0, Graphics.LEFT | Graphics.TOP);
		
		// If there are any non-"color-to-check-for" pixels, we are touching
		return !imageIsColor(i, color_to_check_for);
	}
		
	static private boolean imageIsColor(Image image, int color) {
		// Checks the image to see if every pixel is the specified color
		// (0 == black, 1 == white).
		// (this function should really belong in the image class)
		
		byte[] data = image.getData();
		int width = image.getWidth();
		int height = image.getHeight();
		byte mask = (byte)255;
		// black pixels have bits that are all set; white ones are all clear
		byte target = (color == 0) ? (byte)255 : (byte)0;
		byte maskedTarget = target;
		int index = 0;
		
		// Each byte across the image contains the pixels
		// for up to eight vertical points.
		// (Fewer if there aren't that many rows left in the image).
		
		// Sweep vertically
		for (int y = 0; y < height; y += 8) {

			// and calculate our masks for this row
			int dif = height - y;
			if (dif >= 8) {
				mask = (byte)255;
			} else {
				// if 'dif' is 3, we are only interested in the 
				// 3 least significant bits.
				mask = (byte)((1 << dif) - 1);
			}
			maskedTarget = (byte)(target & mask);
			
			// Sweep horizontally
			for (int x = 0; x < width; ++x) {				
				// and look at up to 8 rows at a time
				if ((data[index++] & mask) != maskedTarget) return false;
			}
		}
		
		return true;
	}

	// This corresponds to the block "if on edge, bounce"
	public void bounceOffEdge() {
		if (!touchingEdge()) return;
		
		// Get our current direction, and break it into component parts
		float radians = (float)Math.toRadians(this.rotationDegrees);
		float xdir = (float)Math.cos(radians);
		float ydir = -(float)Math.sin(radians);
		
		int xadjust = 0;
		int yadjust = 0;
		
		Rectangle rect = getBounds();	// in screen space
		if (rect.x < 0) { 
			xdir = +Math.abs(xdir); 
			xadjust = 0 - rect.x;
		}
		if (rect.x + rect.width >= LCD.SCREEN_WIDTH) { 
			xdir = -Math.abs(xdir); 
			xadjust = (LCD.SCREEN_WIDTH - (rect.x + rect.width)) - 1;
		}
		if (rect.y < 0) { 
			ydir = -Math.abs(ydir); 
			yadjust = - (0 - rect.y);
		}
		if (rect.y + rect.height >= LCD.SCREEN_HEIGHT) { 
			ydir = +Math.abs(ydir); 
			yadjust = -(LCD.SCREEN_HEIGHT - (rect.y + rect.height)) + 1;
		}
		
		// set to degrees based on ydir, xdir, rounded to 3 places.
		this.setRotationDegrees(
			1000 * Math.round(Math.toDegrees(Math.atan2(-ydir, xdir))) / 1000.0f
		);
		
		incrementX(xadjust, false);
		incrementY(yadjust, false);
		drawPenLines();
	}
	
	static float quantizeNonNumericValue(float input) {
		// If the input is NaN, return 0
		// If it is infinite, return a large positive or negative number
		// Otherwise, return the number, rounded to the nearest .00001
		if (Float.isNaN(input)) {
			return 0;
		} else if (Float.isInfinite(input)) {
			if (input > 0.0f) {
				return 1000000.0f;
			} else {
				return -1000000.0f;
			}
		} else {
			return Math.round((double)input * 100000) / 100000.0f;
			//return input;
		}
	}
	public void setPosition(float x, float y) {
		setX(x, false);
		setY(y, false);
		drawPenLines();
	}
	public void setX(float x, boolean drawLines) {
		float newX = quantizeNonNumericValue(x);
		if ((int)newX != (int)this.x) {
			if (drawLines) { drawPenLines(); }
			need_to_redraw = true;
		}
		this.x = newX;	
	}
	public void setY(float y, boolean drawLines) {
		float newY = quantizeNonNumericValue(y);
		if ((int)newY != (int)this.y) {
			if (drawLines) { drawPenLines(); }
			need_to_redraw = true;
		}
		this.y = newY;
	}
	public float getX() {
		return this.x;
	}
	public float getY() {
		return this.y;
	}
	public void incrementX(float dx, boolean drawLines) {
		setX(getX() + dx, drawLines);
	}
	public void incrementY(float dy, boolean drawLines) {	
		setY(getY() + dy, drawLines);
	}
	
	// ROTATION
	public float heading() {
		float h = this.rotationDegrees + 90;
		if (h > 180.0f) { h = h - 360.0f; }
		return h;
	}
	public void setHeading(float h) {
		setRotationDegrees(h - 90);
	}
	public float rotationDegrees() {
		return this.rotationDegrees;
	}
	public void setRotationDegrees(float rotationDegrees) {
		float oldRot = this.rotationDegrees;
		this.rotationDegrees = quantizeNonNumericValue(rotationDegrees) % 360.0f;
		if (this.rotationDegrees != oldRot) {
			need_to_redraw = true;
		}
	}
	public void rotate(float degreesToTurn) {
		setRotationDegrees(rotationDegrees + degreesToTurn);
	}
	public void pointTowards(EncSprite targetSprite) {
		if (targetSprite != null) {
			float dx = targetSprite.getX() - getX();
			float dy = targetSprite.getY() - getY();
			float angle = ((float)Math.toDegrees(Math.atan2(dy, dx)) + 360.0f) % 360.0f;
			setRotationDegrees(angle);
		}
	}
	
	// MOVEMENT
	public void forward(float distance) {
		// move this sprite forward the specified distance (in pixels)
		// avoid NaN or infinite coordinates
		float radians = (float)Math.toRadians(this.rotationDegrees);
		float dX = (float)Math.cos(radians) * distance;
		float dY = -(float)Math.sin(radians) * distance;
		incrementX(dX, false);
		incrementY(dY, false);
		drawPenLines();
		// keepOnScreen();
	}
	public void moveTo(EncSprite targetSprite) {
		if (targetSprite != null) {
			setPosition(targetSprite.getX(), targetSprite.getY());
		}
	}
	public void glideTo(float destX, float destY, float timeInSeconds) {
		// this is a straightforward, naive implementation, and is likely inefficient
		float startX = getX();
		float startY = getY();
		float deltaX = destX - startX;
		float deltaY = destY - startY;
		long start_time = System.currentTimeMillis();
		long end_time = start_time + (int)(1000.0f * timeInSeconds);
		long now;
		
		
		while (true) {
			try { Thread.sleep(0); } catch (InterruptedException e) { }
			now = System.currentTimeMillis();
			if (now >= end_time) break;
			
			// u is the proportion along the motion, in the range [0, 1]
			float u = (float)(now - start_time) / 1000.0f / timeInSeconds;
			setPosition(startX + u * deltaX, startY + u * deltaY);
		 }
		 
		 setPosition(destX, destY); 
	}

	// LOOKS COMMANDS
	public void say(String newMessage) {
		if (!message.equals(newMessage)) {
			message = newMessage;
			need_to_redraw = true;
		}
	}

	public void sayFor(String newMessage, float duration) {
		say(newMessage);
		try {
           Thread.sleep((int)(1000.0f * duration));
       	} catch (Exception ex){}
		say("");       	
	}

	public void hide() {
		if (!hidden) {
			hidden = true;
			need_to_redraw = true;
		}
	}
	public void show() {
		if (hidden) {
			hidden = false;
			need_to_redraw = true;
		}
	}

	// PEN COMMANDS
	public void stamp() {
		if (hidden) return;
	
		int spriteX = X_ORIGIN + (int)x;
		int spriteY = Y_ORIGIN - (int)y;

		Graphics sg = getStageGraphics();
		
		if (sg != null && this.validData) {
			costume.drawOn(sg, spriteX, spriteY);
		}
	}
	public void clear() {
		if (this != stage) {
			stage.clear();
			return;
		}
		int desiredCostume = currentCostume;
		currentCostume = 0;	// do this so the change will actually take effect
		wearCostume(desiredCostume);
	}
	public void penDown(boolean down) {
		if (penDown == down) return;	// ignore duplicate commands

		int spriteX = X_ORIGIN + (int)x;
		int spriteY = Y_ORIGIN - (int)y;

		if (down) {
			prevX = spriteX;
			prevY = spriteY;
		} else {
			drawPenLines();
		}

		penDown = down;
	}
	public void drawPenLines() {
		if (penDown) {
			int spriteX = X_ORIGIN + (int)x;
			int spriteY = Y_ORIGIN - (int)y;
			if (spriteX != prevX || spriteY != prevY) {
				Graphics g = getStageGraphics();
				if (g != null) {
					g.setColor(penColour);
					g.drawLine(prevX, prevY, spriteX, spriteY);
					prevX = spriteX; prevY = spriteY;
					need_to_redraw = true;	// more accurately, it is the stage that needs redrawn
				}
			}
		}
	}	
	
	// may not actually implement this one that uses a number to specify the color	
	public void setPenColor(int clr) {
		if (clr > 50) {
			penColour = 0x00FFFFFF;	// white
		} else {
			penColour = 0x00000000; // black
		}
	}
	
	// takes a color list and returns -1 == invalid, 0 == black, 1 == white
	private int colorListToNumber(EObject HSVlist) {
		if (HSVlist == null) { return -1; }
		HSV color = HSVlist.asHSV();
		if (!color.isValid()) { return -1; }
		return (color.val > 35.0f) ? 1 : 0;
	} 
	
	public void setPenColor(EObject HSVlist) {
		int clrnum = colorListToNumber(HSVlist);
		if (clrnum == 0) penColour = 0x00000000;
		if (clrnum == 1) penColour = 0x00FFFFFF;
	}
}