package enchanting;

import java.util.ArrayList;
import lejos.nxt.comm.LCP;

public final class Variable extends EObject {
    
    protected String fullName;
    protected boolean changed;	// allows you to track if the variable has changed recently
    
	// UNDEFINED
	//
    public Variable(String aFullName) {
    	super();
        fullName = aFullName;
        changed = true;
    }
    
    public synchronized void setValue(EObject value) { setValue(value, true); }    

    public synchronized void setValue(EObject value, boolean sendUpdates) { 
    	if (!equals(value)) {
    		changed = true;
  			super.setValue(value);
			if (sendUpdates) sendUpdate();
		}
	}
    
    // NUMBERS
    //
    public Variable(String aFullName, float value) {
    	this(aFullName);
    	setValue(value);
    }
    public Variable(String aFullName, int value) {
    	this(aFullName);
    	setValue((float)value);
    }
    public Variable(String aFullName, double value) {
    	this(aFullName);
    	setValue((float)value);
    }
    
    @Override
    public synchronized void setValue(float value) { setValue(value, true); }
    @Override
    public synchronized void setValue(int value) { setValue(value, true); }
    @Override
    public synchronized void setValue(double value) { setValue(value, true); }
    
    public synchronized void setValue(float value, boolean sendUpdates) {
		if (type != ETYPE.NUMBER || ( type == ETYPE.NUMBER && asNumber() != value))
			changed = true;
		super.setValue(value);
		if (sendUpdates)
			sendUpdate();
	}
	public synchronized void setValue(int value, boolean sendUpdates) {
		setValue((float)value, sendUpdates);
	}
	public synchronized void setValue(double value, boolean sendUpdates) {
		setValue((double)value, sendUpdates);
	}
    
    // STRING
    //
    public Variable(String aFullName, String initialValue) {
    	this(aFullName);
    	setValue(initialValue);
    } 
    public synchronized void setValue(String value, boolean sendUpdates) {
    	// TODO check for number / boolean inside string 
    	if (type != ETYPE.TEXT || ( type == ETYPE.TEXT && asText() != value ))
    		changed = true;
    	super.setValue(value);
    	if (sendUpdates)
    		sendUpdate();
    }
    @Override
    public synchronized void setValue(String value) { setValue(value, true); } 
    
    // BOOLEAN
    public Variable(String aFullName, boolean initialValue) {
    	this(aFullName);
    	setValue(initialValue);
    }
    public synchronized void setValue(boolean value, boolean sendUpdates) {
    	if (type != ETYPE.BOOLEAN || ( type == ETYPE.BOOLEAN && asBoolean() != value ))
    		changed = true;
    	super.setValue(value);
    	if (sendUpdates)
    		sendUpdate();
    }
    @Override
    public synchronized void setValue(boolean value) { setValue(value, true); } 
    
    // LIST - TODO check for changes when elements are accessed
    public Variable(String aFullName, ArrayList<EObject> initialValue) {
    	this(aFullName);
		setValue(initialValue);
    }
    public synchronized void setValue(ArrayList<EObject> value, boolean sendUpdates) {
    	changed = true;
    	super.setValue(value);
    	if (sendUpdates)
    		sendUpdate();
    }
    @Override
    public synchronized void setValue(ArrayList<EObject> value) { setValue(value, true); } 
    
    // COMMANDS
    public Variable(String aFullName, Command initialValue) {
    	this(aFullName);
    	setValue(initialValue);
    }
    public synchronized void setValue(Command value, boolean sendUpdates) {
    	changed = true;
    	super.setValue(value);
    	if (sendUpdates)
    		sendUpdate();
    }
    @Override
    public synchronized void setValue(Command value) { setValue(value, true); }
      
    // REPORTERS
    public Variable(String aFullName, Reporter initialValue) {
    	this(aFullName);
    	setValue(initialValue);
    }
    public synchronized void setValue(Reporter value, boolean sendUpdates) {
    	changed = true;
    	super.setValue(value);
    	if (sendUpdates)
    		sendUpdate();
    }
    @Override
    public synchronized void setValue(Reporter value) { setValue(value, true); } 
      
    // PREDICATES
    public Variable(String aFullName, Predicate initialValue) {
    	this(aFullName);
    	setValue(initialValue);
    }
    public synchronized void setValue(Predicate value, boolean sendUpdates) {
    	changed = true;
    	super.setValue(value);
    	if (sendUpdates)
    		sendUpdate();
    }
    @Override
    public synchronized void setValue(Predicate value) { setValue(value, true); } 
    
    
    // Updating
    @Override
    public void sendUpdate() {
    	// notify computer that variable has changed
    	String value;
    	if (isList()) 	{ value = toSmalltalkString(); }
		else			{ value = asText(); }
        LCP.updateMsg((byte) 1, fullName + " ", fullName + " \"" + value + "\"");  
        clearChanges(); // JM added this, should it be here?
    }
    
    public boolean hasChanged() {
        return changed;
    }
    
    public void clearChanges() {
        changed = false;   
    }
    
    public String getFullName() {
        return fullName;
    }
    
    // special functions
//     @Override
//     public boolean increment(float inc) {
//     	boolean changed = super.increment(inc);
//     	if (changed) 
//     		sendUpdate();
//     	return changed;
//     }

}